# History: 04/24/2012  Initial coding. 
#          04/27/2012  Add option for a Wald test
#    
###################################################################
#
# NOTES: In the control list for rpart, the cp value should be set to
#          a number less than -1 to ensure the entire tree is grown.
#        This is because, the deviance measure using geeglm does
#        not guarantee that 
#
#        (1)   deviance(parent) >= deviance(left) + deviance(right)
#
#        If the above inequality is not satisfied, then RPART will
#        assign a value of -1 for the complexity parameter (alpha).
#
#        The input y matrix into rpart can contain more columns than
#        just the response columns. See global.offset and global.infoCols.
#
# Global Variables:
#
# Some global variables have default values if they are not specified.
#
# global.test           1 or 2. 1 = likelihood ratio, 2 = Wald test
#                       The default value is 1
# global.data           A data frame containing the X, Z, G variables.
#                       No default
# global.offset.col     Variable for the offset in global.data.
#                       No default 
# global.Z.cols         Variables names for the Z heterogeneous covariates in global.data
#                       No default.
# global.G.cols         Variables for the G genotype variables in global.data
#                       No default.
# global.controls       Vector of TRUE/FALSE or row numbers in global.data that denote the controls
#                       No default
# global.minbucket      Initialize to minbucket option in rpart. 
#                       Default is 10.        
# global.splitMat       Initialize to a matrix with N rows and 2 columns.
###################################################################

# Function to return the dimensions of an object. Return vector is
#   c(# of rows, # of columns)
# A vector x will have 1 column
getDim <- function(x) {
  ret <- dim(as.matrix(x))
  if (is.null(ret)) ret <- c(length(x), 1)
  ret
}

# Function to return a list of all partitions for 2-5 categories
getPartitions <- function(z, ncat) {
  # z       Vector of unique categories
  # ncat    Number of unique categories

  # Hard code the different splits 
  if (ncat == 2) {
    ret <- list(
           list(left=z[1], right=z[2])
           )
  } else if (ncat == 3){
    ret <- list(
           list(left=z[1], right=c(z[2],z[3])), 
           list(left=z[2], right=c(z[1],z[3])),
           list(left=z[3], right=c(z[1],z[2])) 
           )
  } else if (ncat == 4) {
    ret <- list( 
           list(left=z[1],              right=c(z[2],z[3],z[4])), 
           list(left=c(z[1],z[2]),      right=c(z[3],z[4])),
           list(left=c(z[1],z[3]),      right=c(z[2],z[4])),   
           list(left=c(z[1],z[4]),      right=c(z[2],z[3])),
           list(left=c(z[1],z[2],z[3]), right=z[4]),        
           list(left=c(z[1],z[2],z[4]), right=z[3]),
           list(left=c(z[1],z[3],z[4]), right=z[2]) 
           )
  } else if (ncat == 5) {
    ret <- list( 
           list(left=z[1],                   right=c(z[2],z[3],z[4],z[5])), 
           list(left=c(z[1],z[2]),           right=c(z[3],z[4],z[5])),
           list(left=c(z[1],z[3]),           right=c(z[2],z[4],z[5])), 
           list(left=c(z[1],z[4]),           right=c(z[2],z[3],z[5])), 
           list(left=c(z[1],z[5]),           right=c(z[2],z[3],z[4])),
           list(left=c(z[1],z[2],z[3]),      right=c(z[4],z[5])),   
           list(left=c(z[1],z[2],z[4]),      right=c(z[3],z[5])),
           list(left=c(z[1],z[2],z[5]),      right=c(z[3],z[4])),   
           list(left=c(z[1],z[3],z[4]),      right=c(z[2],z[5])),
           list(left=c(z[1],z[3],z[5]),      right=c(z[2],z[4])),  
           list(left=c(z[1],z[4],z[5]),      right=c(z[2],z[3])),
           list(left=c(z[1],z[2],z[3],z[4]), right=z[5]),        
           list(left=c(z[1],z[2],z[3],z[5]), right=z[4]),
           list(left=c(z[1],z[2],z[4],z[5]), right=z[3]),      
           list(left=c(z[1],z[3],z[4],z[5]), right=z[2])
           )
  } else {
    print(ncat)
    print(z)
    stop('Error in getPartition')
  }

  ret
}

# Function to perform a likelihood ratio or Wald test 
vglm_test.wald <- function(data, vars, test.vars, response, offset=NULL) {

  ## if offset is a vector, it will be repeated ##
  if (!is.null(offset)) {
    full.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data[, vars], offset=offset), silent=TRUE)
  } else {
    full.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data[, vars]), silent=TRUE)
  }
  if (global.DEBUG) print(full.model)
  if ("try-error" %in% class(full.model)) return(0)

  parms <- full.model@coefficients
  cov   <- vcov(full.model)

  # Get the correct variables to test
  n <- ncol(full.model@predictors)
  tvars <- paste(test.vars, ":", 1:n, sep="")

  temp  <- waldTest.main(parms, cov, tvars)
  ret   <- temp$test  

  ret

} # END: vglm_test.wald

# Function to perform a likelihood ratio or Wald test 
vglm_test.llr <- function(data, vars, test.vars, response, offset=NULL) {

  ## if offset is a vector, it will be repeated ##
  if (!is.null(offset)) {
    full.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data[, vars], offset=offset), silent=TRUE)
  } else {
    full.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data[, vars]), silent=TRUE)
  }
  if (global.DEBUG) print(full.model)
  if ("try-error" %in% class(full.model)) return(0)

  # Reduced model
  red.vars <- vars[!(vars %in% test.vars)]
  if (!is.null(offset)) {
    red.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data.frame(data[, red.vars]), offset=offset), silent=TRUE)
  } else {
    red.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data.frame(data[, red.vars])), silent=TRUE)
  }
  if (global.DEBUG) print(red.model)
  if ("try-error" %in% class(red.model)) return(0)

  ll1 <- full.model@criterion$loglikelihood
  ll2 <- red.model@criterion$loglikelihood
  ret <- -2*(ll2 - ll1)

  ret

} # END: vglm_test.llr


# Function to compare two coefficient. for the moment, the test.vars has only one variable
vglm_contrast <- function(data, model.vars, test.vars, response, offset=NULL, test=1) {

  # data          Data frame
  # model.vars    Vector of variable names for all variables in the (full) model
  # test.vars     Vector of variable names to test
  # response      Response vector
  # offset        NULL or vector for an offset
  # test          1 or 2   1=likelihood ratio, 2=Wald test

  vars <- unique(model.vars)

  if (!is.null(offset)) {
    full.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data[, vars], offset=offset), silent=TRUE)
  } else {
    full.model <- try(vglm(response ~ -1 + ., family=multinomial, data=data[, vars]), silent=TRUE)
  }
  if (global.DEBUG) print(full.model)
  if ("try-error" %in% class(full.model))  stop("ERROR: vglm_contrast")

  parms <- full.model@coefficients
  cov   <- vcov(full.model)

  # Get the correct variables to test
  n <- ncol(full.model@predictors)
  if (n!= 2)
  {
     stop("ERROR: vglm_contrast")
  }
 
  n.snp<-length(test.vars)
  name.list<-c(test.vars, test.vars)
  num.list<-rep(c(1:2), c(n.snp, n.snp))

  tvars <- paste(name.list, ":", num.list, sep="")
  cov.loc<-cov[tvars, tvars]

  t.1<-parms[tvars[1]] - parms[tvars[2]]
  v.1<-cov.loc[1,1] + cov.loc[2,2] - 2.0 * cov.loc[1,2]
  
  ret <- t.1 * t.1 / v.1

  ret


} # END: vglm_test


# Function to perform a likelihood ratio or Wald test 
vglm_test <- function(data, model.vars, test.vars, response, offset=NULL, test=1) {

  # data          Data frame
  # model.vars    Vector of variable names for all variables in the (full) model
  # test.vars     Vector of variable names to test
  # response      Response vector
  # offset        NULL or vector for an offset
  # test          1 or 2   1=likelihood ratio, 2=Wald test

  vars <- unique(model.vars)
  if (test == 1) {
    ret <- vglm_test.llr(data, vars, test.vars, response, offset=offset)
  } else {
    ret <- vglm_test.wald(data, vars, test.vars, response, offset=offset)
  }

  ret

} # END: vglm_test

# Function to compute the deviance using vglm
nodeDeviance.vglm <- function(dataObj) {
  # This function is called by temp2.
  # temp2 is for determining the best split variable and where to split.
   
  # dataObj      List containing the response and data frame for vglm  

  my.y<-dataObj$response
  my.y[my.y==3]<-0
  ret<-GroupSplit(my.y, dataObj$full.data[, global.G.cols])
 
  # Keep track of the maximum goodness of split value. This will be the
  #  deviance for the parent node.
  if (ret > global.maxSplit) global.maxSplit <<- ret

  ret
}

# The 'evaluation' function.  Called once per node.
#  Produce a label (1 or more elements long) for labeling each node,
#  and a deviance.  The latter is
#	- of length 1
#       - equal to 0 if the node is "pure" in some sense (unsplittable)
#       - does not need to be a devian ce: any measure that gets larger
#            as the node is less acceptable is fine.
#       - the measure underlies cost-complexity pruning, however
temp1 <- function(y, wt, parms) {

  # We have to be tricky here and use global variables because of the
  # way rpart works. For example, rpart will call temp1 to compute the
  # deviance for all the data (top level node). At this time, we don't know 
  # what the deviance should be, because it is based on the geeglm results 
  # from the best split variable. So we let the
  # deviance be 9999 (or any number) and keep track of this node and
  # correct deviance using the global variables global.id, global.splitId,
  # and global.splitMat.
  # The actual deviance for each node is computed later in the temp2
  # function. For now, we set the node deviance to be 9999 and change
  # the deviances in editRpart.

  if (global.DEBUG) print("### Begin temp1 ###")

  # Get the correct global id number
  global.id <<- global.id + 1

  # If global.maxSplit > 0, then we know that we have the deviance
  # value (global.maxSplit) for the parent of the current node.
  if (global.maxSplit > 0) {
    # Update global.splitId
    global.splitId <<- global.splitId + 1

    # Make sure that the global.splitMat has enough rows in it. If not,
    # then add another row.
    if (global.splitId > dim(global.splitMat)[1]) {
      global.splitMat <<- rbind(global.splitMat, global.splitMat[1, ])
      warning("The number of rows in global.splitMat has been exceeded.")
    }

    # Save the maximum value of the goodness of split and id number.
    global.splitMat[global.splitId,] <<- c(global.maxSplit, global.id)
  }
  
  # Reset global.maxSplit
  global.maxSplit <<- -1

  if (global.DEBUG) print("### End temp1 ###")

  # Return list. The returned label must be a scalar.
  list(label=global.id, deviance=9999)
}

# Function to return the response and data frames for the vglm function
getData <- function(yvec, xcat=NULL) {

  # yvec    Input y vector (row numbers of global.data) from the temp2 splitting function   
  # xcat    Vector of TRUE/FALSE of a given split. xcat[i] is TRUE if
  #         observation i has a category in the left partition

  # Define the response, with all the controls at the end
  response <- rep.int(2, length(yvec))
  response[xcat] <- 1
  response <- c(response, rep(3, global.ncontrols))

  # Get the rows to use
  rows <- c(yvec, global.controls)
    
  offset <- global.data[rows, global.offset.col]
  n <- ncol(offset)
  if ((!is.null(n)) && (n == 1)) dim(offset) <- NULL
  
  # Get the data frames
  dat1 <- global.data[rows, c(global.Z.cols, global.G.cols)] 
  dat2 <- global.data[rows, global.Z.cols]
  if (is.null(dim(dat2))) dim(dat2) <- c(length(rows), 1) 

  list(response=response, full.data=dat1, reduced.data=dat2, offset=offset)

}

# The split function, where most of the work occurs.
#   Called once per split variable per node.
# If continuous=T
#   The actual x variable is ordered
#   y is supplied in the sort order of x, with no missings,
#   return two vectors of length (n-1):
#      goodness = goodness of the split, larger numbers are better.
#                 0 = couldn't find any worthwhile split
#        the ith value of goodness evaluates splitting obs 1:i vs (i+1):n
#      direction= -1 = send "y< cutpoint" to the left side of the tree
#                  1 = send "y< cutpoint" to the right
#         this is not a big deal, but making larger "mean y's" move towards
#         the right of the tree, as we do here, seems to make it easier to
#         read
# If continuous=F, x is a set of integers defining the groups for an
#   unordered predictor.  In this case:
#       direction = a vector of length m= "# groups".  It asserts that the
#           best split can be found by lining the groups up in this order
#           and going from left to right, so that only m-1 splits need to
#           be evaluated rather than 2^(m-1)
#       goodness = m-1 values, as before.
#
# The reason for returning a vector of goodness is that the C routine
#   enforces the "minbucket" constraint. It selects the best return value
#   that is not too close to an edge.
temp2 <- function(y, wt, x, parms, continuous) {
    if (global.DEBUG) print("### Begin temp2 ###")

    # Get the dimensions of y
    nrowCol <- getDim(y)
    nobs    <- nrowCol[1]
 
    # For now x will be categorical. continuous should be FALSE
    
	  # Categorical x variable
      # x will have integer values

      # Get the categories
	  ux <- sort(unique(x))

      # Get the number of categories
      ncat <- length(ux)

      # Get all partitions
      partitions <- getPartitions(ux, ncat)

      # Calculate the number of partitions
      npartitions <- length(partitions)

      # Initialize 
      maxGoodness  <- 1e-100
      maxPartition <- 0
        
      # Loop over the different partitions
      for (k in 1:npartitions) {
        # Get the left node observations. temp is a boolean vector of length nobs.
        temp <- (x %in% partitions[[k]]$left) 

        # Check that there are observations in both left and right nodes,
        #  and that they contain at least minbucket observations
        sum <- sum(temp)
        if ((sum < global.minbucket) || (nobs-sum < global.minbucket)) {
          # The left or right node would contain too few observations, so
          # do not consider a split here. Set goodness to 0.
          goodness <- 0
        } else {
          # We have enough observations in both left and right nodes.
          # Get the data frame of offset, Z and G covariates for the vglm function
          dataObj <- getData(y, xcat=temp)

          # Get the goodness of split. 
          goodness <- nodeDeviance.vglm(dataObj)
 
          if ((is.na(goodness)) || (goodness < 0)) goodness <- 0

          # Keep track of the maximum goodness value and the partition
          if (goodness > maxGoodness) {
            maxGoodness  <- goodness
            maxPartition <- k
          }
        } # END: else
  
      } # END: for (k in 1:npartitions)

      # Set all the elements in the goodness vector to 0 except one element
      # that corresponds to maxGoodness.
      # The goodness vector must have length ncat - 1.
      goodness <- double(ncat-1)

      if (maxPartition) {
        # Get the partition of best split
        temp <- partitions[[maxPartition]]
        
        # Set the direction
        ux <- c(temp$left, temp$right)

        # Set the goodness value
        goodness[length(temp$left)] <- maxGoodness
      }

	  retList <- list(goodness=goodness, direction=ux)
      if (global.DEBUG) {
        print(retList)  
        print("### End temp2 ###")
      }

  retList
}
	
# Functions to determine if a number is even or odd
is.even <- function(x) {
  as.logical(1 - (x %% 2))
}

# Function to remove all global variables
global.cleanup <- function(objList=NULL) {
  # objList  Is currently not being used

  # Remove all global variables
  rm(global.minbucket, global.id, global.splitId, global.maxSplit,  
     global.nycols, global.offset.col,
     global.Z.cols, global.G.cols, 
     global.controls, global.ncontrols, global.DEBUG,
     global.splitMat, global.vglmOptions, global.test, 
     inherits=TRUE)
}

# Function to check and initialize all global variables each time rpart is called
global.initialize <- function(y, nrowCol) {
  # y         Response matrix
  # nrowCol   Dimension of y

  # Check for global variables not specified
  if (!exists('global.data')) stop("global.data must be set") 
  if (!exists('global.Z.cols')) stop("global.Z.cols must be set")
  if (!exists('global.G.cols')) stop("global.G.cols must be set")
  if (!exists('global.controls')) stop("global.controls must be set")
  if (!exists('global.offset.col')) stop("global.offset.col must be set")
  global.data <<- as.data.frame(global.data, stringsAsFactors=FALSE)

  if (!exists('global.DEBUG')) global.DEBUG <<- 0

  # Assign default values if the user does not specify them
  if (!exists('global.minbucket')) global.minbucket <<- 30
  if (!exists('global.test')) global.test <<- 1

  # Make sure that minbucket is not too small
  if (global.minbucket < 2) global.minbucket <<- 2
 
  # global.id must be set to 0 each time rpart is called   
  global.id <<- 0

  # splitId and maxSplit must be set each call to rpart
  global.splitId  <<- 0
  global.maxSplit <<- -1

  # If global.controls is logical, convert to integers
  if (is.logical(global.controls)) global.controls <<- as.integer((1:nrow(global.data))[global.controls])
  global.ncontrols <<- length(global.controls) 

  # Get the number of y columns
  global.nycols <<- 1

  # Check global.splitMat. If this matrix does not exist or does not
  #  have 2 columns, then the matrix is created.
  flag <- 0
  if (!exists('global.splitMat')) {
    flag <- 1
  } else {
    if (ncol(global.splitMat) != 2) flag <- 1
  }
  # Create matrix (if necessary)
  if (flag) global.splitMat <<- matrix(nrow=100, ncol=2, data=-1)

  # The column names must be "split" and "id" 
  colnames(global.splitMat) <<- c("split", "id")
 
}

# The init function:
#   fix up y to deal with offsets
#   return a dummy parms list
#   numresp is the number of values produced by the eval routine's "label"
#   numy is the number of columns for y
#   summary is a function used to print one line in summary.rpart
#   text is a function used to put text on the plot in text.rpart
# In general, this function would also check for bad data, see rpart.poisson
#   for instance.
temp3 <- function(y, offset, parms, wt) {

    # Get the dimensions of y
    nrowCol <- getDim(y)

    # Check and initialize global variables
    global.initialize(y, nrowCol)

    # numresp must be set to 1. This is a bug in rpart.
    list(y=y, parms=0, numresp=1, numy=nrowCol[2],
	      summary= function(yval, dev, wt, ylevel, digits ) {
		  paste("  mean=", format(signif(yval, digits)),
			", MSE=" , format(signif(dev/wt, digits)),
			sep='')
	     },
    	 text= function(yval, dev, wt, ylevel, digits, n, use.n ) {
	     if(use.n) {paste(format(yval,digits),"\nn=", n,sep="")}
             else{paste(format(yval,digits))}
           })

}

# Function to edit the return list from rpart. This function returns the
#   same list with the 'frame' field changed.
editRpart <- function(rpartList) {
  # rpartList    The returned list from rpart 

  # Update the deviance and complexity values

  # Get the named item 'frame' in the list
  frame <- rpartList$frame

  # Get the id numbers of the nodes
  ids <- frame[, "yval"]  

  # Update the deviance column for non-leaf nodes
  if (global.splitId > 0) {
    # Get the subset of the global split matrix
    temp <- global.splitMat[1:global.splitId, ] 

    # Watch out if temp only contains 1 row
    if (global.splitId == 1) {
      dim(temp) <- c(1, length(temp))
      colnames(temp) <- colnames(global.splitMat)
    }

    # Get the node numbers in the frame
    nodes <- as.numeric(rownames(frame))

    # Get the frame node numbers of the id column in the split matrix
    nodeNums <- nodes[match(temp[, "id"], ids)]
    nodeNums <- nodeNums[!is.na(nodeNums)]

    if (length(nodeNums) > 0) {
      # We only need the even numbered nodes
      rows     <- is.even(nodeNums)
      nodeNums <- nodeNums[rows] 

      # Get the parent node numbers
      nodeNums <- nodeNums/2

      # Update the frame
      frame[match(nodeNums, nodes), "dev"] <- temp[rows, "split"]
    }
  } # END: if (global.splitId > 0)

  # Set the deviance for leaf nodes to 0
  rows <- frame[, "var"]=="<leaf>"
  frame[rows, "dev"] <- 0

  # Change the complexity parameter to a positive number, otherwise
  #  little info will be printed when summary is called.
  frame[, "complexity"] <- 9999

  # Do a final error check
  if (9999 %in% frame[, "dev"]) stop("ERROR: editRpart frame")

  # Put the edited frame back in the list
  rpartList$frame <- frame

  rpartList
}

# Function to compute the Wald test 
waldTest.main <- function(parms, cov, parmNames) {

  # parms      Parameter vector
  # cov        Covariance matrix
  # parmNames  Character or numeric vector of parameters to test

  df     <- length(parmNames)
  nrcov  <- nrow(cov)
  vnames <- names(parms)

  if (is.numeric(parmNames)) {
    temp <- parmNames %in% 1:nrcov
    vpos <- parmNames[temp]
    np   <- length(vpos)
    if (!np) return(list(test=NA, df=0, pvalue=NA))

    # Update parms and cov
    parms <- parms[vpos]
    cov   <- cov[vpos, vpos] 

  } else {
    # Remove missing names in parmNames
    vnames <- vnames[vnames %in% parmNames]
     
    # Update the parameter vector (the name is kept if length(vnames) = 1)
    parms <- parms[vnames]

    # Remove missing values
    temp   <- !is.na(parms)
    parms  <- parms[temp]
    vnames <- vnames[temp]

    # Check for error
    np <- length(parms)
    if (!np) return(list(test=NA, df=0, pvalue=NA))
    
    # Update cov
    cov <- cov[vnames, vnames]
  }

  if (np == 1) {
    test   <- parms * parms /cov
    if (test >= 0) {
       pvalue   <- pchisq(test, df=1, lower.tail=FALSE) 
    } else {
    pvalue   <- NA
    }   
    return(list(test=test, df=np, pvalue=pvalue))
  } 

  # See if matrix is invertible
  temp <- try(solve(cov), silent=TRUE)
  if (class(temp) == "try-error") {
    return(list(test=NA, df=np, pvalue=NA))
  } 

  # Get the test statistic
  dim(parms) <- c(np, 1)
  test       <- t(parms) %*% temp %*% parms
  dim(test)  <- NULL
  if (test >= 0) {
    pvalue   <- pchisq(test, df=np, lower.tail=FALSE) 
  } else {
    pvalue   <- NA
  }

  list(test=test, df=np, pvalue=pvalue)

} # END: waldTest.main



##Y: outcome, coded as 0, 1, 2, ..., K, where 0 is control group
##G: genotype, coded as 0, 1, 2
GlobalTest<-function(Y, G){
  
  N<-length(Y)
  K<-max(Y)
  stat<-0
  dof<-0
  tmp<-.C("GlobalTest", as.integer(Y), as.integer(K), as.integer(N), as.integer(G), stat=as.double(stat), dof=as.integer(dof))
  stat<-tmp$stat
  dof<-tmp$dof
  stat
  
}

##Y: outcome, coded as 0, 1, 2, ..., K, where 0 is control group
##G: genotype, coded as 0, 1, 2
GroupMerge<-function(Y, G){
  
  N<-length(Y)
  K<-max(Y)
  idx1<-rep(0, K*(K-1)/2)
  idx2<-rep(0, K*(K-1)/2)
  stat<-rep(0, K*(K-1)/2)
  tmp<-.C("GroupMerge_trend", as.integer(Y), as.integer(K), as.integer(N), as.integer(G), stat = as.double(stat), idx1 = as.integer(idx1), idx2 = as.integer(idx2))
  idx1<-tmp$idx1
  idx2<-tmp$idx2
  stat<-tmp$stat
  my.id<-sort(x=stat, decreasing=FALSE, index.return=TRUE)$ix[1]
  c(idx1[my.id], idx2[my.id])  
  
}

##Y: outcome, coded as 0, 1, 2, where 0 is control group. NOte that there are only two case groups when calling this function
##G: genotype, coded as 0, 1, 2
GroupSplit<-function(Y, G){
  
  N<-length(Y)
  if(max(Y) != 2){
    stop("outcome must be coded as 0, 1, 2")
  }
  stat<-0
  tmp<-.C("GroupSplit_trend", as.integer(Y), as.integer(N), as.integer(G), stat=as.double(stat))
  stat<-tmp$stat
  stat
 
}
